/*
##################################################################################################################
# (c) Copyright 2016-2020 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates and licensors (“Micro Focus”)
# are as may be set forth in the express warranty statements accompanying such products and services. Nothing
# herein should be construed as constituting an additional warranty. Micro Focus shall not be liable for technical
# or editorial errors or omissions contained herein. The information contained herein is subject to change without
# notice.
##################################################################################################################
# Source module name: ciplatform.h
##################################################################################################################
# $Author: SPritamPal $              $Locker:  $
# $Revision: $            $Name:  $ $State: Exp $
# $Last Updated Date: 2020/04/27 16:23:15 $SShivani
##################################################################################################################
*/

#ifndef CI_PLATFORM
#define CI_PLATFORM

#define BUFFER_SIZE 256
#define PLATFORM_CONFIG_FILE "/etc/opt/novell/iprint/conf/iprintconf.properties"
#define PLATFORM_IPRINT "platform"
#define PLATFORM_APPLIANCE "appliance"
#define PLATFORM_OES "oes"
#define MOBILITY_ADDON_INSTALLED "mobility_addon_installed"
#define MOBILITY_ADDON_CONFIGURED "mobility_addon_configured"
#define SERVER_LOG_LEVEL "server_log_level"
#define MAX_OPEN_NO_FILES "max_open_files"
#define CLIENT_DELETE_NOT_FOUND_PRINTERS "client_delete_not_found_printers"
#define CURL_TIMEOUT_MS "curl_timeout"
#define DRIVERSTORE_WORKERS_COUNT "driverstore_workers_count"
#define JOB_EXPIRY_TIMEOUT "job_expiry_timeout"
#define STR_TRUE "true"
#define STR_FALSE "false"
#define LOG_LEVEL_WARNING "warning"
#define LOG_LEVEL_INFO "info"
#define JOB_INTERVAL_MINIMUM_TIMEOUT 300	//05min * 60 sec = 300sec
#define JOB_INTERVAL_DEFAULT_TIMEOUT 900	//15min * 60 sec = 900sec
#define JOB_EXPIRY_MAXIMUM_TIMEOUT 8640		//6days * 24hrs * 60min = 8640min

// Note: Initial idea was to include ids.h for below macros.
//       This lead to build issues due to chain of dependencies pulled by ids.h.
//       To resolve build issue, these macros have been moved from ids.h to here.
#define IDS_MAX_NUM_THREADS         100
#define IDS_MIN_NUM_THREADS         5

typedef struct
{
	char platform[BUFFER_SIZE];
	char mobility_addon_installed[16];
	char mobility_addon_configured[16];
	char server_log_level[16];
	short delete_not_found_printers;
	int  countDriverStoreWorkers;
} PlatformConfig;

PlatformConfig platformConfig;
uint  jobExpiryTimeOutGbl;
uint  jobExpiryTimeIntervalGbl;

short IsMobilityFeatureConfigured();
short IsMobilityFeatureInstalled();
short IsOESPlatform();
short IsAppliancePlatform();
short IsJobExpiryFeatureConfigured();
short IsDeleteNotFoundPrintersConfigured();

void ReadPlatformInformationFromFile();
void getServerLogLevel(char* level);

char * trim (char *s);
int ReadCurlTimeout();
int GetDriverStoreWorkersCount();

#endif //CI_PLATFORM

