/*
##################################################################################################################
# (c) Copyright 2001-2020 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates and licensors (“Micro Focus”)
# are as may be set forth in the express warranty statements accompanying such products and services. Nothing
# herein should be construed as constituting an additional warranty. Micro Focus shall not be liable for technical
# or editorial errors or omissions contained herein. The information contained herein is subject to change without
# notice.
##################################################################################################################
# Source module name: acctg.c
##################################################################################################################
*/
//+==========================================================================+
//| Source module name: acctg.c                                              |
//+==========================================================================+
//|$Author: dtaylor $              $Locker:  $
//|$Revision: 1.11 $            $Name:  $ $State: Exp $
//|$Date: 2008/03/13 23:41:39 $
//+===========================================================================

/* Standard Library Header Includes */
   #include <sys/types.h>
   #include <sys/stat.h>
   #include <fcntl.h>
   #include <stdio.h>
   #include <stdlib.h>

   /* __USE_GNU enables strcasestr() */
   #define __USE_GNU
   #include <string.h>

   #include <cimalloc.h>
   #include <nwdpxfer.h>

/* Local Header Includes */
   #define ACCTOID_DEF
      #include "acctoidi.h"
   #undef ACCTOID_DEF
   #define ACCTG_C
      #include "globals.h"
   #undef ACCTG_C

   #include "asutil.h"
   #include "ndpsop_c.h"
   #include "psmops.h"
   #include "threadpl.h"

/* Local Manifest Constant Definitions */
   #define READ_BUFFER_SIZE         1024 * 64

/* Local Macro Definitions */

/* Type Definitions */

/* Structure Definitions */

/* External Function Prototypes */

/* Local Function Prototypes */
   int main( int argc, char *argv[] );
   void WaitForModuleTermination( void );
   void MySignalHandler(int signal);
   void MySetSignalHandler(int signal, void(*sigHandler)(int signal));

   int CheckParentPid(int ppid);

   void GetDisplayTimeString(
      time_t          timeValue,
      char           *timePtr,
      nuint           includeScondsFlag);

   void RemoveCommas(uint8_t *bufferPtr);

   int PrintJob(
      uint8_t                   *printerNamePtr,
      uint8_t                   *jobNamePtr,
      uint8_t                   *jobOwnerPtr,
      uint8_t                   *jobFilePathPtr);

/* Modularly Global Variables */
   int               moduleDownMGbl;
   int               mainPidMGbl;
   int               parentPidMGbl;
   int               parentIsPsmFlagMGbl;
   Process          *mainProcPtrMGbl;

   uint8_t *acctgReportTypeStrPtrMGbl[] =
      {"", "New Job", "Job Created", "Job Processing", "Job Completed"};

   uint8_t *jobCompletionCodeStrPtrMGbl[] =
   {
      "Unknown Job Completion State",
      "Job Completed Successfully",
      "Cancelled by User",
      "Cancelled by Operator",
      "Aborted by Printer"
   };

/*
//+==========================================================================+
// Function Name: main
//+==========================================================================+
// Description:   Load, initialize NLM, and become toDoBroker thread
//+==========================================================================+
// Returns: Nothing.
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int main( int argc, char *argv[] )
{
   int               ccode;
   uint              recLength;
   uint8_t          *tmpStr1Ptr;
   uint8_t          *tmpStr2Ptr;
   uint8_t           buffer[128];

   /*
   // Initialize our global variables
   */
   nlmStateGbl = NLM_NOT_STARTED;

   mainPidMGbl = getpid();
   parentPidMGbl = getppid();
   parentIsPsmFlagMGbl = CheckParentPid(parentPidMGbl);

   /* Parse to the program name in the load string */
   tmpStr1Ptr = argv[0];
   for (tmpStr2Ptr = strchr(tmpStr1Ptr, '/');
        tmpStr2Ptr != NULL;
        tmpStr2Ptr = strchr(tmpStr1Ptr, '/'))
   {
      tmpStr1Ptr = tmpStr2Ptr + 1;
   }

   if (strlen(tmpStr1Ptr) > (sizeof(acctgServiceNameGbl) - 1))
   {
      printf("\nThe name of this program cannot be longer than %lu bytes\n",
             sizeof(acctgServiceNameGbl) - 1);
      goto Exit;
   }
   strcpy(acctgServiceNameGbl, tmpStr1Ptr);

   ccode = CiInterfaceInitMac();
   if (ccode < N_SUCCESS)
      goto Exit;

   CiMutexCreateMac(&taskThreadListMutexGbl);
   CiMutexCreateMac(&processListMutexGbl);

   mainProcPtrMGbl = AcctInitProcessList();
   if(mainProcPtrMGbl == NULL)
      goto Exit;

   /*
   // Let's get to work
   */
   ccode = ThreadPoolInit();

   if (ccode != N_SUCCESS)
      goto Exit;

   nlmStateGbl = NLM_RUNNING;

   logFhGbl = open(ACCTG_LOG_FILENAME, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
   if (logFhGbl != -1)
      lseek(logFhGbl, 0, SEEK_END);

   recLength = sprintf(buffer, "ModuleName=>%s<\n", acctgServiceNameGbl);
   write(logFhGbl, buffer, recLength);

   CiMutexCreateMac (&acctgMutexGbl);
   InitAcctgPsmClientInterface();

   /*
   // Does not return from this call until we are shutting down
   */
   WaitForModuleTermination();
   TermAcctgPsmClientInterface();

   CiMutexDestroyMac (&acctgMutexGbl);

Exit:

   nlmStateGbl = NLM_SHUTTING_DOWN;     /* just to be absolutely sure */

   if (logFhGbl != -1)
      close(logFhGbl);

   /*
   // Deinitialize and shut down
   */
   ThreadPoolTerm();

   AcctFreeProcessList();

   CiInterfaceTermMac();

   if(CiMutexDestroyMac(&processListMutexGbl) == EBUSY)
   {
      CiMutexLockMac(&processListMutexGbl);
      CiMutexUnlockMac(&processListMutexGbl);
      CiMutexDestroyMac(&processListMutexGbl);
   }

   if(CiMutexDestroyMac(&taskThreadListMutexGbl) == EBUSY)
   {
      CiMutexLockMac(&taskThreadListMutexGbl);
      CiMutexUnlockMac(&taskThreadListMutexGbl);
      CiMutexDestroyMac(&taskThreadListMutexGbl);
   }


   exit(0);
}

void MyInitiateModuleShutdown( void )
{
//   raise(SIGTERM);
   kill(mainPidMGbl, SIGTERM);
}

void WaitForModuleTermination( void )
{
   sigset_t set, zeromask;

   sigemptyset(&zeromask);
   sigemptyset(&set);
   sigaddset(&set, SIGSEGV);
   sigaddset(&set, SIGINT);
   sigaddset(&set, SIGTERM);
   sigaddset(&set, SIGKILL);
   // sigaddset(&set, SIGHUP);
   // sigaddset(&set, SIGUSR2);
   sigprocmask(SIG_UNBLOCK, &set, NULL);

   MySetSignalHandler(SIGINT,    MySignalHandler);
   MySetSignalHandler(SIGTERM,   MySignalHandler);
   MySetSignalHandler(SIGHUP,    MySignalHandler);
   MySetSignalHandler(SIGUSR1,   MySignalHandler);
   MySetSignalHandler(SIGCHLD,   MySignalHandler);

   MySetSignalHandler(SIGSEGV, MySignalHandler);
   MySetSignalHandler(SIGBUS, MySignalHandler);
   MySetSignalHandler(SIGFPE, MySignalHandler);
   MySetSignalHandler(SIGILL, MySignalHandler);
   MySetSignalHandler(SIGXCPU, MySignalHandler);
   MySetSignalHandler(SIGXFSZ, MySignalHandler);

//    MySetSignalHandler(SIGKILL, MySignalHandler);
//    MySetSignalHandler(SIGTERM, MySignalHandler);
//    MySetSignalHandler(SIGUSR2, MySignalHandler);
// //    MySetSignalHandler(SIGHUP,  MySignalHandler);

   sigemptyset( &set );
   while (moduleDownMGbl == FALSE)
      // sigsuspend(&zeromask);
      CiSleepMillisecsMac(100);
}
void MySetSignalHandler(
   int   signal,
   void(* sigHandler)(int))
{
   struct sigaction sigStruct;

   memset(&sigStruct, '\0', sizeof(sigStruct));
   sigStruct.sa_handler = sigHandler;
   sigemptyset(&sigStruct.sa_mask);
   sigStruct.sa_flags = 0;
   if (signal != SIGALRM)
      sigStruct.sa_flags |= SA_RESTART;
   sigaction(signal, &sigStruct, NULL);
}

void MySignalHandler(int signal)
{
   switch(signal)
   {
      case SIGINT:
         break;

      case SIGSEGV:
         // kill(getpid(), SIGSTOP);
         // break;
      case SIGTERM:
      case SIGKILL:
      case SIGHUP:
      case SIGUSR2:
         nlmStateGbl = NLM_SHUTTING_DOWN;
         /*
         // If the Print Manager launched this module and is unloading, we tend
         // to time out waiting for a reply if we call the manager.
         // 1. In this case, it signals this module with SIGUSR2.
         //    It will get signalled as this module unloads.
         // 2. Otherwise, we need to unbind so the Manager knows that the
         //       module is gone.
         */
         // if (parentIsPsmFlagMGbl == FALSE)
         // if (signal != SIGUSR2)
         //    TermAcctgPsmClientInterface();
         moduleDownMGbl = TRUE;
         break;
      default:
         break;
   }
}

/*
//+==========================================================================
//| Function Name: CheckParentPid
//+--------------------------------------------------------------------------
//| Description: Check whether the parent process is the PSM.
//+--------------------------------------------------------------------------
//| Returns:
//|           True if comman-line for the parent process contains the PSM
//|               executable name
//|           False if not.
//+--------------------------------------------------------------------------
//| Assumptions: None.
//+==========================================================================
*/
int CheckParentPid(
   int   ppid)
{
   int ccode;
   int cmdLineFH;
   char buffer[PATH_MAX];

   ccode = FALSE;

   /*
   // WARNING!!  NOT PORTABLE!!
   // Need to brute force this.  We'll look in the proc filesystem under the
   //    ppid and see what is contained in the cmdline file.
   // If it's the PSM daemon we'll return TRUE, otherwise FALSE.
   */
   sprintf(buffer, "/proc/%d/cmdline", ppid);
   cmdLineFH = open(buffer, O_RDONLY);
   if(cmdLineFH != -1)
   {
      if (read(cmdLineFH, buffer, sizeof(buffer)) > 0)
      {
         if (strstr(buffer, PSM_DAEMON_NAME) != NULL)
         {
            /*
            // The parent process has the PSM name in its command-line.
            // So we will assume that this is it.
            */
            ccode = TRUE;
         }
      }
      close(cmdLineFH);
   }

   return(ccode);
}

/*
//+==========================================================================+
// Function Name: AccountingBindPa
//+==========================================================================+
// Description:
//+==========================================================================+
// Returns:      0 - N_SUCCESS
//+==========================================================================+
// Assumptions: The opResultStruct is already initialized with Null bytes.
//+==========================================================================+
*/
void AccountingBindPa(
   uint8_t             *requestArgPtr,
   pOpResultStruct      opResultStructPtr)
{
   pAcctgBindPAArg      requestPtr;
   pAcctgBindPAReply    resultPtr;

   requestPtr = (pAcctgBindPAArg)requestArgPtr;
   resultPtr = (pAcctgBindPAReply)opResultStructPtr;

   resultPtr->paHandle = requestPtr->paHandle;
   if (strcasestr(requestPtr->configString, "newjob") != NULL)
      resultPtr->jobAcctgPointMask |= ACCOUNTING_POINT_NEW_JOB;
   if (strcasestr(requestPtr->configString, "jobcreated") != NULL)
      resultPtr->jobAcctgPointMask |= ACCOUNTING_POINT_JOB_CREATED;
   if (strcasestr(requestPtr->configString, "jobstart") != NULL)
      resultPtr->jobAcctgPointMask |= ACCOUNTING_POINT_JOB_START;
   if (strcasestr(requestPtr->configString, "jobclose") != NULL)
      resultPtr->jobAcctgPointMask |= ACCOUNTING_POINT_JOB_CLOSE;
}
   
/*
//+==========================================================================+
// Function Name: AccountingUnbindPa
//+==========================================================================+
// Description:
//+==========================================================================+
// Returns:      0 - N_SUCCESS
//+==========================================================================+
// Assumptions: The opResultStruct is already initialized with Null bytes.
//+==========================================================================+
*/
void AccountingUnbindPa(
   uint8_t             *requestArgPtr,
   pOpResultStruct      opResultStructPtr)
{
}
   
/*
//+==========================================================================+
// Function Name: AccountingProcessInfo
//+==========================================================================+
// Description:
//+==========================================================================+
// Returns:      NWDPSimpleResult
//+==========================================================================+
// Assumptions: The opResultStruct is already initialized with Null bytes.
//+==========================================================================+
*/
int AccountingProcessInfo(
   pAcctgInfoArg              requestBufPtr,
   pOpResultStruct            opResultStructPtr)
{
   // AttrSetAVPStruct           srcAvp;
   NWDPAttributeValue         tmpAttributeValue1;
   NWDPAttributeValue         tmpAttributeValue2;
   NWDPAttributeSet           tmpAttributeSet;
   NWDPObjectIdentification   objectIdentification;
   NWDPObjectIdentifierSet    objectIdentifierSet;
   NWDPSelector               selector;
   int                        errorCode;
   int                        index1;
   int                        index2;
   uint                       bytesWritten;
   uint                       priority;
   uint                       recLength;
   NWDPSessionHandle          sessionHandle;
   pAcctgInfoReply            resultPtr;
   pNWDPAttribute             tmpAttrPtr;
   pNWDPAttributeValue        tmpAttrValuePtr;
   pNWDPListObjectAttrsResult loaResultPtr;
   pNWDPObjectResult          objectResultPtr;
   uint8_t                    submitTimeString[80];
   uint8_t                    completetimeString[80];
   uint8_t                    buffer[1024];
   NWDPAttribute              tmpAttrSet[2];
   NWDPObjectIdentifier       tmpOidSet[2];
   uint8_t                    dummyOid1[] = {6,5,1,2,3,4,5};
   uint8_t                    dummyOid2[] = {6,5,1,2,3,4,6};

   static int                 sendJobFlag = 0;
   errorCode = N_SUCCESS;
   priority = 0;
   resultPtr = &opResultStructPtr->infoReply;

   resultPtr->paHandle = requestBufPtr->paHandle;
   resultPtr->jobNumber = requestBufPtr->jobNumber;
   // resultPtr->jobPageCount = 0;

   if (logFhGbl != -1   &&
       requestBufPtr->acctgReportType == NWDP_ACCTG_ARG_JOB_CLOSE)
   {
      GetDisplayTimeString(requestBufPtr->createTime, submitTimeString, N_FALSE);
      GetDisplayTimeString(requestBufPtr->completeTime, completetimeString, N_FALSE);
      RemoveCommas(submitTimeString);
      RemoveCommas(completetimeString);
      RemoveCommas(requestBufPtr->jobOwner);
      RemoveCommas(requestBufPtr->jobName);
      RemoveCommas(requestBufPtr->printerName);

      recLength = sprintf(buffer,
                          "%s,%s,%s,%s,%u,%s,%s,%u,%u,%u,%s\n",
                          acctgReportTypeStrPtrMGbl[requestBufPtr->acctgReportType],
                          requestBufPtr->jobOwner,
                          requestBufPtr->jobName,
                          requestBufPtr->printerName,
                          (uint)requestBufPtr->jobNumber,
                          submitTimeString,
                          completetimeString,
                          (uint)requestBufPtr->jobPageCount,
                          (uint)requestBufPtr->jobSize,
                          (uint)priority,
                          jobCompletionCodeStrPtrMGbl[requestBufPtr->jobCompleteType]);
      /*
      // Write out our string.
      */
      CiMutexLockMac(&acctgMutexGbl);
      lseek(logFhGbl, 0, SEEK_END);
      bytesWritten = write(logFhGbl, buffer, recLength);
      if(bytesWritten != recLength)
         resultPtr->errorCode = N_FAILURE;
      CiMutexUnlockMac(&acctgMutexGbl);

      /*
      // Example code to Get / Set Psm MOD attributes.
      */
      errorCode = BindToPsm(NWDP_BindPSM,
                            N_TRUE,
                            0,
                            NULL,
                            NULL,
                            NULL,
                            &sessionHandle,
                            NULL, NULL);
      if (errorCode == N_SUCCESS)
      {
         /*
         // Get attribute(s)
         */
         objectIdentification.designator = NWDP_OBJ_ID_SIMPLE_NAME;
         objectIdentification.u.simpleName.itemPtr =
            requestBufPtr->printerName;
         objectIdentification.u.simpleName.itemCount =
            strlen(objectIdentification.u.simpleName.itemPtr) + 1;

         memset(&selector, '\0', sizeof(selector));
         selector.objectIdentificationSeqOption.itemCount = 1;
         selector.objectIdentificationSeqOption.itemPtr   = &objectIdentification;

         tmpOidSet[0].oidStructPtr = (pNWDPOid)&dummyOid1;
         tmpOidSet[0].oidStructSize = sizeof(dummyOid1);
         tmpOidSet[1].oidStructPtr = (pNWDPOid)&dummyOid2;
         tmpOidSet[1].oidStructSize = sizeof(dummyOid2);
         objectIdentifierSet.itemCount = 1;
         objectIdentifierSet.itemPtr = tmpOidSet;
         errorCode = GetObjectAttributes(sessionHandle,
                                         NWDP_LIST_OP_ATTRIBUTES,
                                         (pNWDPOid)&ndpsOCPrinterGbl,
                                         0,
                                         &selector,
                                         &objectIdentifierSet,
                                         NULL,
                                         &loaResultPtr);
         if (errorCode == 0 && loaResultPtr != NULL)
            errorCode = loaResultPtr->errorCode;
         if(errorCode == N_SUCCESS)
         {
            objectResultPtr = loaResultPtr->resultSet.itemPtr;

            for (index1 = loaResultPtr->resultSet.itemCount;
                 index1;
                 index1--, objectResultPtr++)
            {
               if (objectResultPtr->objectIdentification.u.simpleName.itemPtr != NULL)
               {
                  /* Oh look, here is the id for the object that we queried!! */
               }

               tmpAttrPtr = objectResultPtr->attributes.itemPtr;
               for (index2 = objectResultPtr->attributes.itemCount;
                    index2;
                    index2--, tmpAttrPtr++)
               {
                  if (!AttrSetOidCmp(tmpAttrPtr->attributeId.oidStructPtr,
                                     (pNWDPOid)dummyOid1)   &&
                      (tmpAttrValuePtr = tmpAttrPtr->valueSet.itemPtr) != NULL)
                  {
                     /* examine the attribute value returned */
                  }

                  if (!AttrSetOidCmp(tmpAttrPtr->attributeId.oidStructPtr,
                                     (pNWDPOid)dummyOid2)   &&
                      (tmpAttrValuePtr = tmpAttrPtr->valueSet.itemPtr) != NULL)
                  {
                     /* examine the attribute value returned */
                  }
               }
            }

            CiFreeMac(loaResultPtr);
         }

         /*
         // Set attribute(s)
         */
         tmpAttributeValue1.designator = NWDP_AVT_TEXT;
         tmpAttributeValue1.u.text.itemPtr = "Accounting test text One";
         tmpAttributeValue1.u.text.itemCount =
            strlen(tmpAttributeValue1.u.text.itemPtr) + 1;
         tmpAttrSet[0].valueSet.itemCount = 1;
         tmpAttrSet[0].valueSet.itemPtr = &tmpAttributeValue1;

         tmpAttributeValue2.designator = NWDP_AVT_TEXT;
         tmpAttributeValue2.u.text.itemPtr = "Accounting test text Two";
         tmpAttributeValue2.u.text.itemCount =
            strlen(tmpAttributeValue2.u.text.itemPtr) + 1;
         tmpAttrSet[1].valueSet.itemCount = 1;
         tmpAttrSet[1].valueSet.itemPtr = &tmpAttributeValue2;
         tmpAttributeSet.itemCount = 1;
         tmpAttributeSet.itemPtr = tmpAttrSet;

         errorCode = SetObjectAttributes(sessionHandle,
                                         (pNWDPOid)&ndpsOCPrinterGbl,
                                         &objectIdentification,
                                         &tmpAttributeSet);
         UnbindFromPsm(sessionHandle);
      }

      /*
      // Set a dummy page count for demo use.
      */
      resultPtr->jobPageCount = logFhGbl;
   }

   if(!sendJobFlag)
   {
      sendJobFlag = 1;
      PrintJob(requestBufPtr->printerName,
               requestBufPtr->jobName,
               requestBufPtr->jobOwner,
               "/opt/novell/iprint/source/acctg/README");
   }
   return (errorCode);
}

/*
//+==========================================================================+
// Function Name: GetDisplayTimeString
//+==========================================================================+
// Description:
//+==========================================================================+
// Returns:      0 - N_SUCCESS
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
void GetDisplayTimeString(
   time_t               timeValue,
   char                *timePtr,
   nuint                includeScondsFlag)
{

   nuint                i;
   uint8_t              lastChar;
   struct tm            tmStruct;

   if(timeValue == 0)
      timeValue = time(NULL);

   localtime_r(&timeValue, &tmStruct);

   strftime(timePtr, 40, "%x %X", &tmStruct);

   /*
   // Remove any leading spaces.
   */
   while(*timePtr == ' ')
   {
      memmove(timePtr, timePtr + 1, strlen(timePtr));
   }

   /*
   // Remove back to back spaces.
   */
   lastChar = 0;
   for(i = 0;
       timePtr[i] != '\0';
       i++)
   {
      if(timePtr[i] == ' ' && lastChar == ' ')
      {
         memmove(&(timePtr[i]), &(timePtr[i+1]), strlen(&(timePtr[i])));
         if(timePtr[i] == '\0')
            return;
      }
      lastChar = timePtr[i];
   }
}

/*
//+==========================================================================+
// Function Name: RemoveCommas
//+==========================================================================+
// Description:
//+==========================================================================+
// Returns: Nothing.
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
void RemoveCommas(
   uint8_t       *bufferPtr)
{
   uint8_t       *tmpPtr;

   if (bufferPtr == NULL)
      return;

   for(tmpPtr = bufferPtr; *tmpPtr != '\0'; tmpPtr++)
   {
      if(*tmpPtr == ',')
         *tmpPtr = ' ';
   }
}

/*
//+==========================================================================+
// Function Name: RemoveCommas
//+==========================================================================+
// Description:
//+==========================================================================+
// Returns: Nothing.
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int PrintJob(
   uint8_t                   *printerNamePtr,
   uint8_t                   *jobNamePtr,
   uint8_t                   *jobOwnerPtr,
   uint8_t                   *jobFilePathPtr)
{
   NWDPAttributeSet           jobAttrSet;
   NWDPTransferDataArgument   transferDataArg;
   NWDPSessionHandle          sessionHandle;
   int                        errorCode;
   int                        sourceFH;
   uint                       jobId;
   pNWDPTransferDataResult    transferDataResultPtr;
   pNWDPPrintResult           printResultPtr;
   NWDPAttribute              jobAttribute[2];
   NWDPAttributeValue         jobValue[2];

   sessionHandle = 0;

   sourceFH = open(jobFilePathPtr, O_RDWR);
   if (sourceFH == -1)
      return(N_FAILURE);

   transferDataArg.data.itemPtr = CiCallocMac(1, READ_BUFFER_SIZE);
   if(transferDataArg.data.itemPtr == NULL)
   {
      errorCode = CiGetLastErrorMac();
      goto PrintJobError;
   }

   errorCode = BindToPsm(NWDP_BindPA,
                         N_TRUE,
                         0,
                         printerNamePtr,
                         NULL,
                         NULL,
                         &sessionHandle,
                         NULL, NULL);
   if(errorCode)
      goto PrintJobError;

   memset(jobValue, 0, sizeof(jobValue));
   jobValue[0].designator = NWDP_AVT_QUALIFIED_NAME;
   jobValue[0].u.qualifiedName.designator = NWDP_QUALIFIED_NAME_SIMPLE;
   jobValue[0].u.qualifiedName.u.simpleName.itemPtr = jobOwnerPtr;
   jobValue[0].u.qualifiedName.u.simpleName.itemCount =
      strlen(jobValue[0].u.qualifiedName.u.simpleName.itemPtr) + 1;
   jobValue[1].designator = NWDP_AVT_SIMPLE_NAME;
   jobValue[1].u.simpleName.itemPtr = jobNamePtr;
   jobValue[1].u.simpleName.itemCount =
      strlen(jobValue[1].u.simpleName.itemPtr) + 1;

   jobAttribute[0].attributeId.oidStructPtr = (pNWDPOid)ndpsAttJobOwnerGbl;
   jobAttribute[0].attributeId.oidStructSize =
         NWDPOidLenMac(jobAttribute[0].attributeId.oidStructPtr);
   jobAttribute[0].valueSet.itemCount = 1;
   jobAttribute[0].valueSet.itemPtr = &jobValue[0];

   jobAttribute[1].attributeId.oidStructPtr = (pNWDPOid)ndpsAttJobNameGbl;
   jobAttribute[1].attributeId.oidStructSize =
         NWDPOidLenMac(jobAttribute[1].attributeId.oidStructPtr);
   jobAttribute[1].valueSet.itemCount = 1;
   jobAttribute[1].valueSet.itemPtr = &jobValue[1];

   jobAttrSet.itemPtr = jobAttribute;
   jobAttrSet.itemCount = sizeof(jobAttribute) / sizeof(jobAttribute[0]);

   errorCode = SubmitPrintJob(sessionHandle,
                              NWDP_PRINT_ARG_CREATE_JOB,
                              &jobId,
                              &jobAttrSet,
                              NULL,
                              NULL);

   transferDataArg.sessionHandle  = sessionHandle;
   transferDataArg.status         = 0;
   for(;;)
   {
      transferDataArg.data.itemCount =
         read(sourceFH, transferDataArg.data.itemPtr, READ_BUFFER_SIZE);
      if(transferDataArg.data.itemCount <=0)
         break;

      errorCode = nwdp_printprogram_a(NWDP_TransferData,
                                      (pnstr)&transferDataArg,
                                      (ppnuint8)&transferDataResultPtr);
      if (errorCode == N_SUCCESS)
      {
         if (transferDataResultPtr->errorCode != 0)
            errorCode = N_FAILURE;

         CiFreeMac(transferDataResultPtr);
      }
      if(errorCode)
         break;
   }

   /*
   // Signal the PA the job has been fully submitted
   // If no data was transferred, the Psm will delete the job.
   */
   errorCode = SubmitPrintJob(sessionHandle,
                              NWDP_PRINT_ARG_CLOSE_JOB,
                              &jobId,
                              NULL,
                              NULL,
                              &printResultPtr);
   if (errorCode == N_SUCCESS)
   {
      /*
      // Free the result struct that we got from the PrintJob() call.
      // Note: the result struct contains Job State info.
      */
      CiFreeMac(printResultPtr);
   }

PrintJobError:
   if(sessionHandle != 0)
      UnbindFromPsm(sessionHandle);
  if(transferDataArg.data.itemPtr != NULL)
      CiFreeMac(transferDataArg.data.itemPtr);
   close(sourceFH);
   return(errorCode);
}
