/*
//+==========================================================================+
//| © Copyright 2001 – 2019 Micro Focus or one of its affiliates.					|
//|																					|
//| The only warranties for products and services of Micro Focus and its affiliates	|
//| and licensors (“Micro Focus”) are set forth in the express warranty statements	|
//| accompanying such products and services. Nothing herein should be construed		|
//| as constituting an additional warranty.											|
//| Micro Focus shall not be liable for technical or editorial errors 				|
//| or omissions contained herein. 													|
//| The information contained herein is subject to change without notice.			|
//| Contains Confidential Information. Except as specifically indicated otherwise,	|
//| a valid license is required for possession, use or copying. 					|
//| Consistent with FAR 12.211 and 12.212, Commercial Computer Software, 			|
//| Computer Software Documentation, and Technical Data for Commercial Items		|
//| are licensed to the U.S. Government under vendor's standard commercial license.	|
//+==========================================================================+
//| Source module name: ndpsop_c.c
//+==========================================================================+
//|$Author: dtaylor $              $Locker:  $
//|$Revision: 1.9 $            $Name:  $ $State: Exp $
//|$Date: 2008/03/13 23:41:39 $
//+===========================================================================
*/
/* Standard Library Header Includes */

/* iPrint Header Includes */
   #include <string.h>

   #include <cimalloc.h>
   #include <nwdpxfer.h>

/* Local Header Includes */
   #include "globals.h"
   #include "ndpsop_c.h"
   #include "psmpackc.h"
   #include "threadpl.h"

/* Local Manifest Constant Definitions */
   #define INTERFACE_STATE_ACTIVE                  1
   #define INTERFACE_STATE_UNINITIALIZED          -1
   #define INTERFACE_STATE_INITIALIZING           -2
   #define INTERFACE_STATE_CONNECTION_LOST        -3
   #define INTERFACE_STATE_OFF                    -4

   #define INTERFACE_INIT_RETRY_SEC_INTERVAL       3
   #define INTERFACE_ACCESS_RETRY_MSEC_INTERVAL  250
   /*
   // 20 * 250 ms = 5 seconds
   */
   #define INTERFACE_ACCESS_RETRY_TIME_LIMIT      20

   #define INTERFACE_INIT_RETRY_MAX_COUNT    40

/* Local Macro Definitions */

/* Type Definitions */
   typedef pnstr (*LocalFnPtr)();
   typedef pXferDataBuf (*PACK_FUNC_PTR)(pnuint8 requestPtr, int *packedRequestSize);
   typedef pnuint8      (*UNPACK_FUNC_PTR)(pnuint8 resultDataPtr);

/* Structure Definitions */

/* External Function Prototypes */

/* Local Function Prototypes */
   void AcctgPsmXferClientStartCommunications(
      void                *dummy);

   int AcctgPsmXferClientCallbackFunc(
      enum CiIpcReqType    request,
      int                  socket,
      CiIpcBuffer_t       *bufPtr,
      int                  bufLen);

/* Modularly Global Variables */

   int      acctgPsmXferClientSocketMGbl = INTERFACE_STATE_UNINITIALIZED;
   int      acctgPsmXferClientSendMutexInitFlagMGbl;
   int      acctgPsmXferClientInterfaceInitCountMGbl;
   // int      acctgPsmXferClientInitRetryCountMGbl = 0;

   CI_THREAD_HANDLE  acctgPsmClientCommThreadHandleMGbl;
   CI_MUTEX          acctgPsmClientXferMutexMGbl;

/*
//+===========================================================================
// Function Name: InitAcctgPsmClientInterface
//+===========================================================================
// Description:
//+===========================================================================
// Returns:      0 - N_SUCCESS
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
int InitAcctgPsmClientInterface(void)
{
   int                  resultCode;

   resultCode = N_SUCCESS;

   if (acctgPsmXferClientSocketMGbl == INTERFACE_STATE_UNINITIALIZED)
   {
      /*
      // Change acctgPsmXferClientSocketMGbl from -1 so other threads won't be
      //    launched.
      */
      acctgPsmXferClientSocketMGbl = INTERFACE_STATE_INITIALIZING;

      /*
      // Get a mutex to prevent multiple threads from sending data
      // at the same time.
      */
      if (acctgPsmXferClientSendMutexInitFlagMGbl == N_FALSE)
      {
         acctgPsmXferClientSendMutexInitFlagMGbl = N_TRUE;
         CiMutexCreateMac(&acctgPsmClientXferMutexMGbl);
      }

      resultCode = CiThreadCreateJoinableHandleMac((CiThreadStartFuncPtr)
                                                      AcctgPsmXferClientStartCommunications,
                                                   NULL,
                                                   &acctgPsmClientCommThreadHandleMGbl);
   }
   return resultCode;
}

/*
//+===========================================================================
// Function Name: TermAcctgPsmClientInterface
//+===========================================================================
// Description:
//+===========================================================================
// Returns:      0 - N_SUCCESS
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
void TermAcctgPsmClientInterface()
{
   int               tmpSocket;

   if (acctgPsmXferClientSocketMGbl == INTERFACE_STATE_UNINITIALIZED)
      return;

   /*
   // Copy acctgPsmXferClientSocketMGbl, then suspend it.
   */
   tmpSocket = acctgPsmXferClientSocketMGbl;
   acctgPsmXferClientSocketMGbl = INTERFACE_STATE_OFF;

   //
   // Cancel the thread handling communications with the PSM.
   //
   CiThreadCancelMac(acctgPsmClientCommThreadHandleMGbl);

   //
   // Join with the thread cancelled above before continuing on.
   //
   CiThreadJoinMac(acctgPsmClientCommThreadHandleMGbl, NULL);

   if (CiMutexDestroyMac(&acctgPsmClientXferMutexMGbl) == EBUSY)
   {
      CiMutexLockMac(&acctgPsmClientXferMutexMGbl);
      CiMutexUnlockMac(&acctgPsmClientXferMutexMGbl);
      CiMutexDestroyMac(&acctgPsmClientXferMutexMGbl);
   }

   if (tmpSocket >= 0)
      close(tmpSocket);
}

/*
//+===========================================================================
// Function Name: AcctgPsmXferClientStartCommunications
//+===========================================================================
// Description:
//+===========================================================================
// Returns: Nothing.
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
void AcctgPsmXferClientStartCommunications(
   void             *dummy)
{
   for (;;)
   {
      acctgPsmXferClientInterfaceInitCountMGbl++;
      CiIpcClientRun(CI_IPC_UDS,
                     CI_IPC_PSM_ACCTG_UDSNAME,
                     0,
                     AcctgPsmXferClientCallbackFunc);

      /*
      // Check to see if we are shutting down
      */
      if (acctgPsmXferClientSocketMGbl == INTERFACE_STATE_OFF ||
          nlmStateGbl == NLM_SHUTTING_DOWN)
      {
         break;
      }

      /*
      // If we successfully connect, 'acctgPsmXferClientInterfaceInitCountMGbl' is
      // reset in AcctgPsmXferClientCallbackFunc
      */
      // if (acctgPsmXferClientInterfaceInitCountMGbl >= INTERFACE_INIT_RETRY_MAX_COUNT)
      //    break;

      sleep(INTERFACE_INIT_RETRY_SEC_INTERVAL);
      //
      // See if we have been cancelled before continuing on.
      //
      CiThreadTestCancelMac();
   }
}

/*
//+===========================================================================
// Function Name: ClientCallBackFunc
//+===========================================================================
// Description:
//+===========================================================================
// Returns:      0 - N_SUCCESS
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
int AcctgPsmXferClientCallbackFunc(
   enum CiIpcReqType request,
   int               socket,
   CiIpcBuffer_t    *bufPtr,
   int               bufLen)
{
   int               errorCode;
   XferDataBuf     *xferDataPtr;

   switch(request)
   {
      case CI_IPC_CONN_NEW:
         // MyMsgPrintf(DEBUG_IPRINT, "XFER_CL_NEW_CONNECT");

         /*
         // We have successfully connected so reset our retry count.
         */
         acctgPsmXferClientInterfaceInitCountMGbl = 0;

         /*
         // Remember the socket in case we want to terminate our
         // connection to the Print Manager
         */
         acctgPsmXferClientSocketMGbl = socket;

         /*
         // Hand the task to the thread pool. The memory associated with
         // bufPtr should be freed up downstream.
         */
         errorCode = ScheduleWorkOnThread((void *)BindAccountingToPsm,
                                          &acctgServiceNameGbl);

         break;

      case CI_IPC_CONN_LOST:
         // MyMsgPrintf(DEBUG_IPRINT, "XFER_CL_CONNECTION_LOST");

         /*
         // We've lost our communications channel to the Print Manager
         // Our life no longer has meaning
         */
         close(acctgPsmXferClientSocketMGbl);
         acctgPsmXferClientSocketMGbl = INTERFACE_STATE_CONNECTION_LOST;

         break;

      case CI_IPC_DATA_READ:
         if(acctgPsmXferClientSocketMGbl == INTERFACE_STATE_OFF)
         {
            /*
            // We are going down.
            // The listening thread is going away.
            */
            CiFreeMac(bufPtr);
         }
         else
         {
            /*
            // Someone is making noise, process it. Put the socket in the
            // prevPtr so that we know who to send the answer to later.
            */
            bufPtr->prevPtr = (void *)((ptrdiff_t)socket);
            xferDataPtr = (pXferDataBuf)&bufPtr->data;

            /*
            // Set the data pointer so the caller can find the data, then
            //  let the calling thread run.
            */
            if (xferDataPtr->opCode == NWDP_PsmCallbackOp)
            {
               *xferDataPtr->dataPtr = (pXferDataBuf)xferDataPtr;
               xferDataPtr->dataLength = bufLen - XFER_DATA_OFFSET;
               // CiMutexLockMac(&acctgPsmClientXferMutexMGbl);
               // CiCondSignalMac(xferDataPtr->condPtr);
               // CiMutexUnlockMac(&acctgPsmClientXferMutexMGbl);
               break;
            }

            /*
            // Hand the task to the thread pool. The memory associated with
            // bufPtr should be freed up downstream.
            */
            errorCode = ScheduleWorkOnThread(acctg_program_a, bufPtr);
            if(errorCode)
            {

            }
         }
         break;

      default:
         /* Should never get here */
         break;
   }

   return(0);
}

/*
//+===========================================================================
// Function Name: nwdp_printprogram_a
//+===========================================================================
// Description:
//+===========================================================================
// Returns:      0 - N_SUCCESS
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
int nwdp_printprogram_a(
   nuint32              opCode,
   pnuint8              requestPtr,
   ppnuint8             resultDataPtr)
{
   // CI_COND              xferReqCond;
   uint                 packedRequestSize;
   int                  errorFlag;
   int                  index1;
   int                  index2;
   pXferDataBuf         packedRequestPtr;
   pXferDataBuf         replyBufPtr;

   PACK_FUNC_PTR        packOp;
   UNPACK_FUNC_PTR      unpackRes;

   packedRequestPtr  = NULL;
   replyBufPtr       = NULL;
   *resultDataPtr    = NULL;

   if(acctgPsmXferClientSocketMGbl == INTERFACE_STATE_UNINITIALIZED)
      goto nwdp_printprogram_aError;

   switch (opCode)
   {
      case NULLPROC:
         errorFlag = N_SUCCESS;
         goto nwdp_printprogram_aEnd;

      case NWDP_BindPSM:
      case NWDP_BindPSM_OAUTH:
         packOp    = (PACK_FUNC_PTR)PsmPackBind;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackBindPsmRes;
         break;

      case NWDP_BindPA:
      case NWDP_BindPA_OAUTH:
         packOp    = (PACK_FUNC_PTR)PsmPackBind;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackBindPaRes;
         break;

      case NWDP_Unbind:
         packOp    = (PACK_FUNC_PTR)PsmPackSimpleArg;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackSimpleRes;
         break;

      case NWDP_Print:
         packOp    = (PACK_FUNC_PTR)PsmPackPrint;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackPrintRes;
         break;

      case NWDP_ListObjectAttributes:
         packOp    = (PACK_FUNC_PTR)PsmPackListAttrs;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackListAttrsRes;
         break;

      case NWDP_Create:
         packOp    = (PACK_FUNC_PTR)PsmPackCreate;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackCreateRes;
         break;

      case NWDP_Delete:
         packOp    = (PACK_FUNC_PTR)PsmPackDelete;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackDeleteRes;
         break;

      case NWDP_Set:
         packOp    = (PACK_FUNC_PTR)PsmPackSet;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackSetRes;
         break;

      case NWDP_TransferData:
         packOp    = (PACK_FUNC_PTR)PsmPackTransferData;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackTransferDataRes;
         break;

      case NWDP_AccountingBindPsm:
         packOp    = (PACK_FUNC_PTR)PsmPackAcctgBindPsm;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackSimpleRes;
         break;

      case NWDP_AccountingUnbindPsm:
         packOp    = (PACK_FUNC_PTR)PsmPackAcctgUnbindPsm;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackSimpleRes;
         break;

      case NWDP_AccountingJobAction:
         packOp    = (PACK_FUNC_PTR)PsmPackAcctgJobAction;
         unpackRes = (UNPACK_FUNC_PTR)PsmUnpackSimpleRes;
         break;

      default:
         errorFlag = NDPS_E_UNSUPPORTED_OP;
         goto nwdp_printprogram_aEnd;
         break;
   }

   packedRequestPtr = packOp(requestPtr, &packedRequestSize);

   if (packedRequestPtr == NULL)
      goto nwdp_printprogram_aError;

   packedRequestPtr->opCode = opCode;
   // packedRequestPtr->condPtr = &xferReqCond;
   packedRequestPtr->dataPtr = &replyBufPtr;

   /*
   // Give the client interface time to initialize.
   */
   errorFlag = NDPS_E_COMMUNICATION;
   for (index1 = INTERFACE_ACCESS_RETRY_TIME_LIMIT; // 5 seconds
        index1;
        index1--)
   {
      if(acctgPsmXferClientSocketMGbl == INTERFACE_STATE_OFF)
         break;

      if (acctgPsmXferClientSocketMGbl >= 0)
      {
         // if (CiCondInitMac(&xferReqCond) == N_SUCCESS)
         // {
            CiMutexLockMac(&acctgPsmClientXferMutexMGbl);
            errorFlag = CiIpcSend(acctgPsmXferClientSocketMGbl,
                                  (pnuint8)packedRequestPtr,
                                  packedRequestSize);

            // if (errorFlag == N_SUCCESS)
            // {
            //    /*
            //    // time out after NWDP_OP_TIMEOUT plus one second.
            //    // unless we are unbinding from the Psm.  This seems to timeout.
            //    */
            //    CiCondTimedWaitMac(&xferReqCond,
            //                       &acctgPsmClientXferMutexMGbl,
            //                       opCode == NWDP_AccountingUnbindPsm
            //                         ? 2
            //                         : NWDP_OP_TIMEOUT + 1);
            // }
            CiMutexUnlockMac( &acctgPsmClientXferMutexMGbl );
            if (errorFlag == N_SUCCESS)
            {
               for(index2 = NWDP_OP_TIMEOUT_MAX_COUNT; index2; index2--)
               {
                  if(replyBufPtr != NULL  ||
                     acctgPsmXferClientSocketMGbl == INTERFACE_STATE_OFF)
                  {
                     break;
                  }
                  CiSleepMillisecsMac(NWDP_OP_TIMEOUT_SLEEP_MS);
               }
            }
         //    CiCondDestroyMac(&xferReqCond);
         // }
         break;
      }

      CiSleepMillisecsMac(INTERFACE_ACCESS_RETRY_MSEC_INTERVAL);
   }

   if (errorFlag != N_SUCCESS || replyBufPtr == NULL)
   {
      errorFlag = NDPS_E_COMMUNICATION;
      goto nwdp_printprogram_aEnd;
   }

   if (replyBufPtr->dataLength == sizeof(nuint32))
   {
      errorFlag = *(pnuint32)&replyBufPtr->dataBuf;
      if (errorFlag != N_SUCCESS)
         goto nwdp_printprogram_aEnd;
   }

   *resultDataPtr = (pnuint8)((*unpackRes)(replyBufPtr->dataBuf));
   if (*resultDataPtr == NULL)
      goto nwdp_printprogram_aError;

   errorFlag = N_SUCCESS;
   goto nwdp_printprogram_aEnd;

nwdp_printprogram_aError:
   errorFlag = N_FAILURE;

nwdp_printprogram_aEnd:
   if (packedRequestPtr != NULL)
      CiFreeMac(packedRequestPtr);
   if (replyBufPtr != NULL)
      FreeIpcResultMac(replyBufPtr);

   return(errorFlag);
}

/*
//+===========================================================================
// Function Name: acctg_program_a
//+===========================================================================
// Description:
//+===========================================================================
// Returns:      0 - N_SUCCESS
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
void acctg_program_a(
   void                *requestBufPtr)
{
   OpResultStruct       opResultStruct;
   int                  packedResultSize;
   int                  index1;
   int                  socket;
   int32                opCode;
   nint32               errorFlag;
   nuint32              opTimeout;
   pXferDataBuf         requestDataPtr;
   pXferDataBuf         packedResultPtr;
   LocalFnPtr           localFnPtr;
   nuint8               errorBuffer[sizeof(XferDataBuf) + sizeof(nuint32)];

   pnuint8              (*packRes)();

   errorFlag       = N_FAILURE;
   packedResultPtr = NULL;

   if (nlmStateGbl == NLM_SHUTTING_DOWN)
   {
      errorFlag = NDPS_E_PSM_GOING_DOWN;
      goto acctg_program_aReply;
   }

   opTimeout = time(NULL) + NWDP_OP_TIMEOUT;

   socket = (int)((ptrdiff_t)((CiIpcBuffer_t *)requestBufPtr)->prevPtr);
   requestDataPtr = (pXferDataBuf)((CiIpcBuffer_t *)requestBufPtr)->data;
   opCode = requestDataPtr->opCode;

   switch (opCode)
   {
      case NULLPROC:
         errorFlag = N_SUCCESS;
         goto acctg_program_aReply;

      case NWDP_AccountingBindPa:
         packRes  = PsmPackAcctgBindPaRes;
         localFnPtr = (LocalFnPtr)AccountingBindPa;
         break;

      case NWDP_AccountingUnbindPa:
         packRes  = PsmPackSimpleRes;
         localFnPtr = (LocalFnPtr)AccountingUnbindPa;
         break;

      case NWDP_SubmitAccountingInfo:
         packRes  = PsmPackAcctgInfoRes;
         localFnPtr = (LocalFnPtr)AccountingProcessInfo;
         break;

      default:
         errorFlag = NDPS_E_UNSUPPORTED_OP;
         goto acctg_program_aReply;
         break;
   }

   memset(&opResultStruct, 0, sizeof(opResultStruct));
   (*localFnPtr)(requestDataPtr->dataBuf, &opResultStruct);

   errorFlag = N_SUCCESS;

      packedResultPtr = (pXferDataBuf)(*packRes)(&opResultStruct, &packedResultSize);

      if (packedResultPtr == NULL)
         errorFlag = NDPS_E_BAD_PARAMETER;

acctg_program_aReply:
   /*
   // If we do not have a normal reply buffer, send a default error buffer.
   */
   if (packedResultPtr == NULL)
   {
      packedResultPtr = (pXferDataBuf)&errorBuffer;
      *(pnuint32)&packedResultPtr->dataBuf = errorFlag;
      packedResultSize = sizeof(errorBuffer);
   }

   packedResultPtr->opCode = NWDP_PsmCallbackOp;
   packedResultPtr->condPtr = requestDataPtr->condPtr;
   packedResultPtr->dataPtr = requestDataPtr->dataPtr;

   CiMutexLockMac(&acctgPsmClientXferMutexMGbl);
   /*
   // Don't send a reply if the client is giving up waiting.
   */
   if (time(NULL) <= opTimeout)
      CiIpcSend(socket, (pnuint8)packedResultPtr, packedResultSize);
   CiMutexUnlockMac(&acctgPsmClientXferMutexMGbl);

   if (requestBufPtr != NULL)
      CiFreeMac(requestBufPtr);

   if (packedResultPtr != (pXferDataBuf)&errorBuffer)
      CiFreeMac(packedResultPtr);
}

