/*
##################################################################################################################
# (c) Copyright 2001-2020 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates and licensors (“Micro Focus”)
# are as may be set forth in the express warranty statements accompanying such products and services. Nothing
# herein should be construed as constituting an additional warranty. Micro Focus shall not be liable for technical
# or editorial errors or omissions contained herein. The information contained herein is subject to change without
# notice.
##################################################################################################################
# Source module name: psmops.c
##################################################################################################################
*/
//+==========================================================================+
//| Source module name: psmops.c
//+==========================================================================+
//|$Author: dtaylor $              $Locker:  $
//|$Revision: 1.4 $            $Name:  $ $State: Exp $
//|$Date: 2008/03/13 23:41:39 $
//+===========================================================================
/* Standard Library Header Includes */
   #include <string.h>

/* Local Header Includes */
   #include <cimalloc.h>

   #include "acctoidi.h"
   #include "psmops.h"

/* Local Manifest Constant Definitions */

/* Local Macro Definitions */

/* Type Definitions */

/* Structure Definitions */

/* External Function Prototypes */

/* Local Function Prototypes */

/* Modularly Global Variables */

/*
//+==========================================================================+
// Function Name: BindToPsm
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int BindToPsm(
   uint                 opCode,
   int                  localBindFlag,
   nuint32              clientAddress,
   pnuint8              printerNamePtr,
   pnuint8              loginNamePtr,
   pnuint8              userNamePtr,
   NWDPSessionHandle   *sessionHandlePtr,
   pnuint8              printerNameOptPtr,
   pAuthAttr            userAuthAttrPtr)
{
   NWDPBindPAArgument   bindArg;
   int                  rCode;
   pNWDPBindPSMResult   resultPtr;

   memset(&bindArg, 0, sizeof(bindArg));

   if (printerNamePtr != NULL)
   {
      bindArg.printerName.itemPtr = printerNamePtr;
      bindArg.printerName.itemCount = strlen(bindArg.printerName.itemPtr) + 1;
   }

   if (loginNamePtr != NULL)
   {
      bindArg.loginName.itemPtr = loginNamePtr;
      bindArg.loginName.itemCount = strlen(bindArg.loginName.itemPtr) + 1;
   }

   if (userNamePtr != NULL)
   {
      bindArg.userName.itemPtr = userNamePtr;
      bindArg.userName.itemCount = strlen(bindArg.userName.itemPtr) + 1;
   }

   if (userAuthAttrPtr != NULL)
   {
      if (userAuthAttrPtr->token.itemPtr != NULL)
      {
         bindArg.isOAuthAuthentication = 1;
		 bindArg.authAttr.token.itemPtr = userAuthAttrPtr->token.itemPtr;
		 bindArg.authAttr.token.itemCount = strlen(bindArg.authAttr.token.itemPtr) + 1;
         if (opCode == NWDP_BindPSM)
            opCode = NWDP_BindPSM_OAUTH;
         if (opCode == NWDP_BindPA)
            opCode = NWDP_BindPA_OAUTH;
      }
   }
   bindArg.clientAddress = clientAddress;
   bindArg.localBindFlag = localBindFlag;

   rCode = nwdp_printprogram_a(opCode,
                               (pnstr)&bindArg,
                               (ppnuint8)&resultPtr);

   if (printerNameOptPtr != NULL &&
       resultPtr != NULL         &&
       resultPtr->psmName.u.simpleName.itemPtr != NULL)
   {
      memcpy(printerNameOptPtr,
             resultPtr->psmName.u.simpleName.itemPtr,
             resultPtr->psmName.u.simpleName.itemCount);
   }

   if (rCode == N_SUCCESS)
   {
      if (resultPtr != NULL)
      {
          rCode = resultPtr->errorCode;
          if (rCode == N_SUCCESS)
             *sessionHandlePtr = resultPtr->sessionHandle;
          CiFreeMac(resultPtr);
      }
   }

   return(rCode);
}

/*
//+==========================================================================+
// Function Name: UnbindFromPsm
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
void UnbindFromPsm(
   NWDPSessionHandle    sessionHandle)
{
   NWDPUnbindArgument   unbindArg;
   pnuint8              resultPtr;

   unbindArg.sessionHandle = sessionHandle;
   if (!nwdp_printprogram_a(NWDP_Unbind, (pnstr)&unbindArg, &resultPtr))
   {
      if (resultPtr != NULL)
         CiFreeMac(resultPtr);
   }
}

/*
//+===========================================================================
// Function Name: BindAccountingToPsm
//+===========================================================================
// Description:
//
//+===========================================================================
// Returns:
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
void BindAccountingToPsm(
   uint8_t                   *accountingServiceNamePtr)
{
   AcctgBindPsmArg            acctgBindPsmArg;
   int                        errorCode;
   pNWDPSimpleResult          acctgBindPsmResultPtr;

   if (strlen(accountingServiceNamePtr) > (sizeof(acctgBindPsmArg.name) - 1))
   {
      CiSetLastErrorMac(CI_ERR_INVALID_PARAM_1);
      return;
   }

   memset(&acctgBindPsmArg, '\0', sizeof(acctgBindPsmArg));

   strcpy(acctgBindPsmArg.name, accountingServiceNamePtr);

   errorCode = nwdp_printprogram_a(NWDP_AccountingBindPsm,
                                   (pnstr)&acctgBindPsmArg,
                                   (ppnuint8)&acctgBindPsmResultPtr);
   if (errorCode == N_SUCCESS)
   {
      errorCode = acctgBindPsmResultPtr->errorCode;
      CiFreeMac(acctgBindPsmResultPtr);
   }
}

/*
//+===========================================================================
// Function Name: UnbindAccountingFromPsm
//+===========================================================================
// Description:
//
//+===========================================================================
// Returns:
//+===========================================================================
// Assumptions: None.
//+===========================================================================
*/
void UnbindAccountingFromPsm(
   uint8_t                   *accountingServiceNamePtr)
{
   AcctgUnbindPsmArg          acctgUnbindPsmArg;
   int                        errorCode;
   pNWDPSimpleResult          acctgUnbindPsmResultPtr;

   if (strlen(accountingServiceNamePtr) > (sizeof(acctgUnbindPsmArg.name) - 1))
   {
      CiSetLastErrorMac(CI_ERR_INVALID_PARAM_1);
      return;
   }

   memset(&acctgUnbindPsmArg, '\0', sizeof(acctgUnbindPsmArg));

   strcpy(acctgUnbindPsmArg.name, accountingServiceNamePtr);

   errorCode = nwdp_printprogram_a(NWDP_AccountingUnbindPsm,
                                   (pnstr)&acctgUnbindPsmArg,
                                   (ppnuint8)&acctgUnbindPsmResultPtr);
   if (errorCode == N_SUCCESS)
   {
      errorCode = acctgUnbindPsmResultPtr->errorCode;
      CiFreeMac(acctgUnbindPsmResultPtr);
   }
}

/*
//+==========================================================================+
// Function Name: AcctgJobAction
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int AcctgJobAction(
   int                  paHandle,
   uint                 jobNumber,
   AcctgJobActionEnum   jobAction)
{
   int                  errorCode;
   AcctgJobActionArg    acctgJobActionArg;
   pNWDPDeleteResult    acctgJobActionResultPtr;

   acctgJobActionArg.paHandle =  paHandle;
   acctgJobActionArg.jobNumber = jobNumber;
   acctgJobActionArg.jobAction = jobAction;

   errorCode = nwdp_printprogram_a(NWDP_AccountingJobAction,
                                  (pnstr)&acctgJobActionArg,
                                  (ppnuint8)&acctgJobActionResultPtr);
   if (errorCode == N_SUCCESS)
   {
      if (acctgJobActionResultPtr != NULL)
      {
         errorCode = acctgJobActionResultPtr->errorCode;
         CiFreeMac(acctgJobActionResultPtr);
      }
   }

   return(errorCode);
}

/*
//+==========================================================================+
// Function Name: CreateObject
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int CreateObject(
   NWDPSessionHandle          sessionHandle,
   pNWDPOid                   classOidPtr,
   pNWDPObjectIdentification  objectIdPtr,
   pNWDPAttributeSet          attrSetPtr,
   int                        forceFlag)
{
   int                        errorCode;
   NWDPCreateArgument         createArg;
   pNWDPCreateResult          createResultPtr;

   memset(&createArg, '\0', sizeof(createArg));

   createArg.sessionHandle = sessionHandle;
   createArg.objectClass.oidStructPtr = classOidPtr;
   createArg.objectClass.oidStructSize =
      NWDPOidLenMac(createArg.objectClass.oidStructPtr);
   createArg.objectId = *objectIdPtr;
   createArg.force = forceFlag;

   if (attrSetPtr != NULL)
      createArg.objectAttrSetOption = *attrSetPtr;

   errorCode = nwdp_printprogram_a(NWDP_Create,
                                  (pnstr)&createArg,
                                  (ppnuint8)&createResultPtr);
   if (errorCode == N_SUCCESS)
   {
      if (createResultPtr != NULL)
      {
         errorCode = createResultPtr->errorCode;
         CiFreeMac(createResultPtr);
      }
   }

   return(errorCode);
}

/*
//+==========================================================================+
// Function Name: DeleteObject
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int DeleteObject(
   NWDPSessionHandle          sessionHandle,
   pNWDPOid                   classOidPtr,
   pNWDPObjectIdentification  objectIdPtr)
{
   int                        errorCode;
   NWDPDeleteArgument         deleteArg;
   pNWDPDeleteResult          deleteResultPtr;

   memset(&deleteArg, '\0', sizeof(deleteArg));

   deleteArg.sessionHandle = sessionHandle;
   deleteArg.objectClass.oidStructPtr = classOidPtr;
   deleteArg.objectClass.oidStructSize =
      NWDPOidLenMac(deleteArg.objectClass.oidStructPtr);
   deleteArg.objectId = *objectIdPtr;

   errorCode = nwdp_printprogram_a(NWDP_Delete,
                                  (pnstr)&deleteArg,
                                  (ppnuint8)&deleteResultPtr);
   if (errorCode == N_SUCCESS)
   {
      if (deleteResultPtr != NULL)
      {
         errorCode = deleteResultPtr->errorCode;
         CiFreeMac(deleteResultPtr);
      }
   }

   return(errorCode);
}

/*
//+==========================================================================+
// Function Name: SetObjectAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int SetObjectAttributes(
   NWDPSessionHandle          sessionHandle,
   pNWDPOid                   classOidPtr,
   pNWDPObjectIdentification  objectIdPtr,
   pNWDPAttributeSet          attrSetPtr)
{
   int                        errorCode;
   NWDPSetArgument            setArg;
   pNWDPSetResult             setResultPtr;

   memset(&setArg, '\0', sizeof(setArg));

   setArg.sessionHandle = sessionHandle;
   setArg.objectClass.oidStructPtr = classOidPtr;
   setArg.objectClass.oidStructSize =
      NWDPOidLenMac(setArg.objectClass.oidStructPtr);
   setArg.objectId = *objectIdPtr;
   setArg.attrModificationSet = *attrSetPtr;

   errorCode = nwdp_printprogram_a(NWDP_Set,
                                  (pnstr)&setArg,
                                  (ppnuint8)&setResultPtr);
   if (errorCode == N_SUCCESS)
   {
      if (setResultPtr != NULL)
      {
         errorCode = setResultPtr->errorCode;
         CiFreeMac(setResultPtr);
      }
   }

   return(errorCode);
}

/*
//+==========================================================================+
// Function Name: GetObjectAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int GetObjectAttributes(
   NWDPSessionHandle             sessionHandle,
   NWDPListOperatorEnum          listOperator,
   pNWDPOid                      classOidPtr,
   nuint32                       scope,
   pNWDPSelector                 selectorOptionPtr,
   pNWDPObjectIdentifierSet      requestedAttrsOptionPtr,
   pnuint8                       listContinuationPtr,
   pNWDPListObjectAttrsResult   *loaResultPtr)
{
   NWDPListObjectAttrsArgument   loaArg;

   memset(&loaArg, '\0', sizeof(loaArg));

   if (listContinuationPtr != NULL)
   {
      loaArg.listAttrsOperation.designator = NWDP_LIST_ATTRIBUTES_ARG_CONT;
      loaArg.listAttrsOperation.u.continuation.context.itemCount =
         sizeof(pnuint8);
      loaArg.listAttrsOperation.u.continuation.context.itemPtr =
         (pnint8)&listContinuationPtr;
   }
   else
   {
      loaArg.listAttrsOperation.designator = NWDP_LIST_ATTRIBUTES_ARG_SPEC;
      loaArg.listAttrsOperation.u.specification.scope = scope;
      loaArg.listAttrsOperation.u.specification.listOperator = listOperator;
      loaArg.listAttrsOperation.u.specification.selectorOptionPtr =
         selectorOptionPtr;

      loaArg.listAttrsOperation.u.specification.objectClass.oidStructPtr =
         classOidPtr;
      loaArg.listAttrsOperation.u.specification.objectClass.oidStructSize =
         NWDPOidLenMac(loaArg.listAttrsOperation.u.specification.
            objectClass.oidStructPtr);

      loaArg.listAttrsOperation.u.specification.requestedAttrsOptionPtr =
         requestedAttrsOptionPtr;
   }

   loaArg.sessionHandle = sessionHandle;
   return(nwdp_printprogram_a(NWDP_ListObjectAttributes,
                              (pnstr)&loaArg,
                              (ppnuint8)loaResultPtr));
}

/*
//+==========================================================================+
// Function Name: GetPrinterAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns: Success: 0
//          Failure: Error code from failed operation.
//+==========================================================================+
// Assumptions: On success, an attribute set is returned that must be freed
//                by the caller.
//+==========================================================================+
*/
int GetPrinterAttributes(
   NWDPSessionHandle             sessionHandle,
   uint8_t                      *printerNamePtr,
   pNWDPObjectIdentifierSet      requestedAttrsOptionPtr,
   pAttrSetStruct               *attrSetPtrPtr)
{
   NWDPObjectIdentification      objectIdentification;
   NWDPSelector                  selector;
   NWDPListObjectAttrsArgument   loaArg;
   AttrSetAVPStruct              objAvp;
   int                           errorCode;
   pNWDPListObjectAttrsResult    loaResultPtr;
   pNWDPObjectResult             objectResultPtr;

   objectIdentification.designator = NWDP_OBJ_ID_SIMPLE_NAME;
   objectIdentification.u.simpleName.itemPtr = printerNamePtr;
   objectIdentification.u.simpleName.itemCount =
      strlen(objectIdentification.u.simpleName.itemPtr) + 1;

   memset(&selector, '\0', sizeof(selector));
   selector.objectIdentificationSeqOption.itemCount = 1;
   selector.objectIdentificationSeqOption.itemPtr   = &objectIdentification;

   memset(&loaArg, '\0', sizeof(loaArg));
   loaArg.listAttrsOperation.designator = NWDP_LIST_OP_ATTRIBUTES;
   loaArg.listAttrsOperation.u.specification.listOperator = NWDP_LIST_OP_ATTRIBUTES;
   loaArg.listAttrsOperation.u.specification.selectorOptionPtr = &selector;

   loaArg.listAttrsOperation.u.specification.objectClass.oidStructPtr =
      (pNWDPOid)&ndpsOCPrinterGbl;
   loaArg.listAttrsOperation.u.specification.objectClass.oidStructSize =
      NWDPOidLenMac(loaArg.listAttrsOperation.u.specification.
         objectClass.oidStructPtr);

   loaArg.listAttrsOperation.u.specification.requestedAttrsOptionPtr =
      requestedAttrsOptionPtr;

   errorCode = nwdp_printprogram_a(NWDP_ListObjectAttributes,
                                   (pnstr)&loaArg,
                                   (ppnuint8)loaResultPtr);
   if (errorCode == N_SUCCESS)
      errorCode = loaResultPtr->errorCode;
   if (errorCode == N_SUCCESS)
   {
      objectResultPtr = loaResultPtr->resultSet.itemPtr;
      if (objectResultPtr != NULL && objectResultPtr->attributes.itemPtr != NULL)
      {
         errorCode = AttrSetCreateRef(objectResultPtr->attributes.itemCount,
                                      attrSetPtrPtr,
                                      NULL);
         if (errorCode == N_SUCCESS)
         {
            objAvp.attrSetRef =
               (pAttrSetStruct)&objectResultPtr->attributes;
            errorCode = AttrSetAppendAttrSet(&objAvp,
                                             &objectResultPtr->attributes);
            if (errorCode != N_SUCCESS)
               AttrSetReleaseRef(attrSetPtrPtr);
         }
      }
      CiFreeMac(loaResultPtr);
   }

   if (errorCode != N_SUCCESS)
      *attrSetPtrPtr = NULL;
   return (errorCode);
}

/*
//+==========================================================================+
// Function Name: GetJobAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns: Success: 0
//          Failure: Error code from failed operation.
//+==========================================================================+
// Assumptions: On success, an attribute set is returned that must be freed
//                by the caller.
//+==========================================================================+
*/
int GetJobAttributes(
   NWDPSessionHandle             sessionHandle,
   uint8_t                      *printerNamePtr,
   uint                          jobNumber,
   pNWDPObjectIdentifierSet      requestedAttrsOptionPtr,
   pAttrSetStruct               *attrSetPtrPtr)
{
   NWDPObjectIdentification      objectIdentification;
   NWDPSelector                  selector;
   NWDPListObjectAttrsArgument   loaArg;
   AttrSetAVPStruct              objAvp;
   int                           errorCode;
   pNWDPListObjectAttrsResult    loaResultPtr;
   pNWDPObjectResult             objectResultPtr;

   objectIdentification.designator = NWDP_OBJ_ID_PRT_CONTAIND_OBJ_ID;
   objectIdentification.u.prtContainedObjectId.printerName.itemPtr =
      printerNamePtr;
   objectIdentification.u.prtContainedObjectId.printerName.itemCount =
      strlen(objectIdentification.u.prtContainedObjectId.printerName.itemPtr) + 1;
   objectIdentification.u.prtContainedObjectId.localIdentifier = jobNumber;

   memset(&selector, '\0', sizeof(selector));
   selector.objectIdentificationSeqOption.itemCount = 1;
   selector.objectIdentificationSeqOption.itemPtr   = &objectIdentification;

   memset(&loaArg, '\0', sizeof(loaArg));
   loaArg.listAttrsOperation.designator = NWDP_LIST_OP_ATTRIBUTES;
   loaArg.listAttrsOperation.u.specification.listOperator = NWDP_LIST_OP_ATTRIBUTES;
   loaArg.listAttrsOperation.u.specification.selectorOptionPtr = &selector;

   loaArg.listAttrsOperation.u.specification.objectClass.oidStructPtr =
      (pNWDPOid)&ndpsOCPrinterGbl;
   loaArg.listAttrsOperation.u.specification.objectClass.oidStructSize =
      NWDPOidLenMac(loaArg.listAttrsOperation.u.specification.
         objectClass.oidStructPtr);

   loaArg.listAttrsOperation.u.specification.requestedAttrsOptionPtr =
      requestedAttrsOptionPtr;

   errorCode = nwdp_printprogram_a(NWDP_ListObjectAttributes,
                                   (pnstr)&loaArg,
                                   (ppnuint8)loaResultPtr);
   if (errorCode == N_SUCCESS)
      errorCode = loaResultPtr->errorCode;
   if (errorCode == N_SUCCESS)
   {
      objectResultPtr = loaResultPtr->resultSet.itemPtr;
      if (objectResultPtr != NULL && objectResultPtr->attributes.itemPtr != NULL)
      {
         errorCode = AttrSetCreateRef(objectResultPtr->attributes.itemCount,
                                      attrSetPtrPtr,
                                      NULL);
         if (errorCode == N_SUCCESS)
         {
            objAvp.attrSetRef =
               (pAttrSetStruct)&objectResultPtr->attributes;
            errorCode = AttrSetAppendAttrSet(&objAvp,
                                             &objectResultPtr->attributes);
            if (errorCode != N_SUCCESS)
               AttrSetReleaseRef(attrSetPtrPtr);
         }
      }
      CiFreeMac(loaResultPtr);
   }

   if (errorCode != N_SUCCESS)
      *attrSetPtrPtr = NULL;
   return (errorCode);
}

/*
//+==========================================================================+
// Function Name: GetDocAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns: Success: 0
//          Failure: Error code from failed operation.
//+==========================================================================+
// Assumptions: On success, an attribute set is returned that must be freed
//                by the caller.
//+==========================================================================+
*/
int GetDocAttributes(
   NWDPSessionHandle             sessionHandle,
   uint8_t                      *printerNamePtr,
   uint                          jobNumber,
   uint                          docNumber,
   pNWDPObjectIdentifierSet      requestedAttrsOptionPtr,
   pAttrSetStruct               *attrSetPtrPtr)
{
   NWDPObjectIdentification      objectIdentification;
   NWDPSelector                  selector;
   NWDPListObjectAttrsArgument   loaArg;
   AttrSetAVPStruct              objAvp;
   int                           errorCode;
   pNWDPListObjectAttrsResult    loaResultPtr;
   pNWDPObjectResult             objectResultPtr;

   objectIdentification.designator = NWDP_OBJ_ID_DOCUMENT_IDENTIFIER;
   objectIdentification.u.documentIdentifier.jobIdentifier.printerName.itemPtr =
      printerNamePtr;
   objectIdentification.u.documentIdentifier.jobIdentifier.printerName.itemCount =
      strlen(objectIdentification.u.documentIdentifier.jobIdentifier.
               printerName.itemPtr) + 1;
   objectIdentification.u.documentIdentifier.jobIdentifier.localIdentifier =
      jobNumber;
   objectIdentification.u.documentIdentifier.documentNumber = docNumber;

   memset(&selector, '\0', sizeof(selector));
   selector.objectIdentificationSeqOption.itemCount = 1;
   selector.objectIdentificationSeqOption.itemPtr   = &objectIdentification;

   memset(&loaArg, '\0', sizeof(loaArg));
   loaArg.listAttrsOperation.designator = NWDP_LIST_OP_ATTRIBUTES;
   loaArg.listAttrsOperation.u.specification.listOperator = NWDP_LIST_OP_ATTRIBUTES;
   loaArg.listAttrsOperation.u.specification.selectorOptionPtr = &selector;

   loaArg.listAttrsOperation.u.specification.objectClass.oidStructPtr =
      (pNWDPOid)&ndpsOCPrinterGbl;
   loaArg.listAttrsOperation.u.specification.objectClass.oidStructSize =
      NWDPOidLenMac(loaArg.listAttrsOperation.u.specification.
         objectClass.oidStructPtr);

   loaArg.listAttrsOperation.u.specification.requestedAttrsOptionPtr =
      requestedAttrsOptionPtr;

   errorCode = nwdp_printprogram_a(NWDP_ListObjectAttributes,
                                   (pnstr)&loaArg,
                                   (ppnuint8)loaResultPtr);
   if (errorCode == N_SUCCESS)
      errorCode = loaResultPtr->errorCode;
   if (errorCode == N_SUCCESS)
   {
      objectResultPtr = loaResultPtr->resultSet.itemPtr;
      if (objectResultPtr != NULL && objectResultPtr->attributes.itemPtr != NULL)
      {
         errorCode = AttrSetCreateRef(objectResultPtr->attributes.itemCount,
                                      attrSetPtrPtr,
                                      NULL);
         if (errorCode == N_SUCCESS)
         {
            objAvp.attrSetRef =
               (pAttrSetStruct)&objectResultPtr->attributes;
            errorCode = AttrSetAppendAttrSet(&objAvp,
                                             &objectResultPtr->attributes);
            if (errorCode != N_SUCCESS)
               AttrSetReleaseRef(attrSetPtrPtr);
         }
      }
      CiFreeMac(loaResultPtr);
   }

   if (errorCode != N_SUCCESS)
      *attrSetPtrPtr = NULL;
   return (errorCode);
}

/*
//+==========================================================================+
// Function Name: SetPrinterAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns: Success: 0
//          Failure: Error code from failed operation.
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int SetPrinterAttributes(
   NWDPSessionHandle             sessionHandle,
   uint8_t                      *printerNamePtr,
   pNWDPAttributeSet             attrSetPtr)
{
   NWDPObjectIdentification      objectIdentification;

   objectIdentification.designator = NWDP_OBJ_ID_SIMPLE_NAME;
   objectIdentification.u.simpleName.itemPtr = printerNamePtr;
   objectIdentification.u.simpleName.itemCount =
      strlen(objectIdentification.u.simpleName.itemPtr) + 1;

   return SetObjectAttributes(sessionHandle,
                              (pNWDPOid)&ndpsOCPrinterGbl,
                              &objectIdentification,
                              attrSetPtr);
}

/*
//+==========================================================================+
// Function Name: SetJobAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns: Success: 0
//          Failure: Error code from failed operation.
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int SetJobAttributes(
   NWDPSessionHandle             sessionHandle,
   uint8_t                      *printerNamePtr,
   uint                          jobNumber,
   pNWDPAttributeSet             attrSetPtr)
{
   NWDPObjectIdentification      objectIdentification;

   objectIdentification.designator = NWDP_OBJ_ID_PRT_CONTAIND_OBJ_ID;
   objectIdentification.u.prtContainedObjectId.printerName.itemPtr =
      printerNamePtr;
   objectIdentification.u.prtContainedObjectId.printerName.itemCount =
      strlen(objectIdentification.u.prtContainedObjectId.printerName.itemPtr) + 1;
   objectIdentification.u.prtContainedObjectId.localIdentifier = jobNumber;

   return SetObjectAttributes(sessionHandle,
                              (pNWDPOid)&ndpsOCPrinterGbl,
                              &objectIdentification,
                              attrSetPtr);
}

/*
//+==========================================================================+
// Function Name: SetDocAttributes
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns: Success: 0
//          Failure: Error code from failed operation.
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int SetDocAttributes(
   NWDPSessionHandle             sessionHandle,
   uint8_t                      *printerNamePtr,
   uint                          jobNumber,
   uint                          docNumber,
   pNWDPAttributeSet             attrSetPtr)
{
   NWDPObjectIdentification      objectIdentification;

   objectIdentification.designator = NWDP_OBJ_ID_DOCUMENT_IDENTIFIER;
   objectIdentification.u.documentIdentifier.jobIdentifier.printerName.itemPtr =
      printerNamePtr;
   objectIdentification.u.documentIdentifier.jobIdentifier.printerName.itemCount =
      strlen(objectIdentification.u.documentIdentifier.jobIdentifier.
               printerName.itemPtr) + 1;
   objectIdentification.u.documentIdentifier.jobIdentifier.localIdentifier =
      jobNumber;
   objectIdentification.u.documentIdentifier.documentNumber = docNumber;

   return SetObjectAttributes(sessionHandle,
                              (pNWDPOid)&ndpsOCPrinterGbl,
                              &objectIdentification,
                              attrSetPtr);
}

/*
//+==========================================================================+
// Function Name: PrintJob
//+==========================================================================+
// Description:
//
//+==========================================================================+
// Returns:
//+==========================================================================+
// Assumptions: None.
//+==========================================================================+
*/
int SubmitPrintJob(
   NWDPSessionHandle          sessionHandle,
   NWDPPrintArgEnum           designator,
   pnuint                     jobIdPtr,
   pNWDPAttributeSet          jobAttributesPtr,
   pNWDPAttributeSet          docAttrsOptionPtr,
   pNWDPPrintResult          *printResultOptionPtr)
{
   int                        errorCode;
   NWDPPrintArgument          printArg;
   NWDPDocumentDescription    docDescription;
   pNWDPPrintResult           myPrintResultPtr;

   memset(&printArg, 0, sizeof(NWDPPrintArgument));

   if (designator != NWDP_PRINT_ARG_CLOSE_JOB)
   {
      memset(&docDescription, 0, sizeof(NWDPDocumentDescription));
      docDescription.transferMethod.oidStructPtr =
            (pNWDPOid)&ndpsValXferMethodNdpsDataMGbl;
      docDescription.transferMethod.oidStructSize =
            NWDPOidLenMac(docDescription.transferMethod.oidStructPtr);
      /*
      docDescription.documentContentOptionPtr = &docContent;
      docContent.designator                   = NWDP_DOC_CONTENT_INCLUDED;
      docContent.u.includedDocument.itemPtr   = jobData;
      docContent.u.includedDocument.itemCount = sizeof(jobData);
      */
      if (docAttrsOptionPtr != NULL)
         docDescription.documentAttributes = *docAttrsOptionPtr;
   }

   printArg.sessionHandle = sessionHandle;
   printArg.printOperation.designator = designator;

   if (designator == NWDP_PRINT_ARG_CREATE_JOB)
   {
      /*
      // The NDPS Manager will default the printerName to the bound printer.
      */
      // printArg.printOperation.u.createJob.printerName            = destPrinterName;
      // printArg.printOperation.u.createJob.jobSubmissionComplete  = N_TRUE;
      printArg.printOperation.u.createJob.jobAttributes = *jobAttributesPtr;
      printArg.printOperation.u.createJob.firstDocumentOptionPtr =
         &docDescription;
   }
   else
   {
      printArg.printOperation.u.addDocument.existingJob.localIdentifier =
         *jobIdPtr;
      printArg.printOperation.u.addDocument.newDocumentPtr = &docDescription;
   }

   errorCode = nwdp_printprogram_a(NWDP_Print,
                                   (pnstr)&printArg,
                                   (ppnuint8)&myPrintResultPtr);
   if (errorCode == N_SUCCESS)
   {
      if (myPrintResultPtr != NULL)
      {
         errorCode = myPrintResultPtr->errorCode;
         if (errorCode == N_SUCCESS)
         {
            *jobIdPtr = myPrintResultPtr->jobIdentification.localIdentifier;
            if (printResultOptionPtr != NULL)
            {
              /*
              // return a pointer to the result struct; the caller must free it.
              */
              *printResultOptionPtr = myPrintResultPtr;
            }
            else
            {
               CiFreeMac(myPrintResultPtr);
            }
         }
      }
   }

   return(errorCode);
}

